﻿Imports System.Windows.Media.Media3D



Public Module M6DOF




    Public Function ToRadians(ByVal degrees As Single) As Single
        Return degrees / (180 / Math.PI)
    End Function

    Public Function ToDegrees(ByVal radians As Single) As Single
        Return radians * (180 / Math.PI)
    End Function










    Public Class Simulator

        Public actuators As New Hashtable

        Public roll As Single = 0
        Public pitch As Single = 0
        Public heave As Single = 0
        Public sway As Single = 0
        Public surge As Single = 0
        Public tl As Single = 0

        Public surge_pitch_component As Single = 0
        Public sway_roll_component As Single = 0


        Public tl_dx As Integer = 800

        Public lever_length As Integer = 120




        Public roll_limit As Integer  '+/- deg
        Public pitch_limit As Integer  '+/- deg

        Public surge_limit As Integer  '+/- mm

        Public sway_limit As Integer '+/- mm

        Public heave_min As Integer  ' mm above motor shafts
        Public heave_max As Integer  ' mm above motor shafts

        Public tl_limit As Integer  ' +/- deg

        Public surge_pitch_limit As Integer  '+/- 
        Public sway_roll_limit As Integer  '+/- deg






        Public Sub New()

            Call UpdateLimits()


            actuators.Add("1", New Actuator("1", 270, 400, -40, 355, 196, 0, 30, lever_length, 485, 90, 270))
            actuators.Add("6", New Actuator("6", 210, 435, -40, 0, 411, 0, 90, lever_length, 485, 90, 270))
            actuators.Add("4", New Actuator("4", -480, -35, -40, -355, -205, 0, -150, lever_length, 485, 90, 270))


            actuators.Add("2", New Actuator("2", 270, -400, -40, 355, -196, 0, -30, lever_length, 485, -90, 90))
            actuators.Add("3", New Actuator("3", 210, -435, -40, 0, -411, 0, -90, lever_length, 485, -90, 90))
            actuators.Add("5", New Actuator("5", -480, 35, -40, -355, 205, 0, 150, lever_length, 485, -90, 90))





        End Sub

        Public Sub UpdateLimits()

            roll_limit = My.Settings.Item("range_roll")
            pitch_limit = My.Settings.Item("range_pitch")
            surge_limit = My.Settings.Item("range_surge")
            sway_limit = My.Settings.Item("range_sway")
            heave_min = My.Settings.Item("range_heave_min")
            heave_max = My.Settings.Item("range_heave_max")
            tl_limit = My.Settings.Item("range_tl")
            surge_pitch_limit = My.Settings.Item("range_surge_pitch")
            sway_roll_limit = My.Settings.Item("range_sway_roll")


        End Sub





        Public Function TwelveBitsToDOF(ByVal twelve_bit_value As Integer, ByVal dof_name As String) As Single

            Dim min_limit As Integer
            Dim max_limit As Integer

            Select Case dof_name
                Case "roll"
                    min_limit = -roll_limit
                    max_limit = roll_limit
                Case "pitch"
                    min_limit = -pitch_limit
                    max_limit = pitch_limit
                Case "surge"
                    min_limit = -surge_limit
                    max_limit = surge_limit
                Case "sway"
                    min_limit = -sway_limit
                    max_limit = sway_limit
                Case "heave"
                    min_limit = heave_min
                    max_limit = heave_max
                Case "tl"
                    min_limit = -tl_limit
                    max_limit = tl_limit
                Case "surge_pitch"
                    min_limit = -surge_pitch_limit
                    max_limit = surge_pitch_limit
                Case "sway_roll"
                    min_limit = -sway_roll_limit
                    max_limit = sway_roll_limit
            End Select


            Dim range As Single = max_limit - min_limit

            Dim center As Single = (max_limit + min_limit) / 2

            Dim offset As Single = (twelve_bit_value - 2048) / 4096

            Return center + range * offset

        End Function







        Public Sub reset()



            For Each act As Actuator In Me.actuators.Values
                act.AttachmentPointReset()
            Next


        End Sub


        Public ReadOnly Property attachment_points As Point3D()
            Get

                Dim ap(5) As Point3D
                Dim i As Integer = 0
                For Each a In actuators.Values
                    ap(i) = CType(a, Actuator).attachment_point
                    i += 1
                Next

                Return ap
            End Get
        End Property





        Public Sub UpdateCoordinates()

            Me.reset()

            ' apply roll
            Dim tf_roll As New RotateTransform3D(New AxisAngleRotation3D(New Vector3D(1, 0, 0), Me.roll + Me.sway_roll_component))

            ' apply pitch
            Dim tf_pitch As New RotateTransform3D(New AxisAngleRotation3D(New Vector3D(0, 1, 0), Me.pitch + Me.surge_pitch_component))

            ' apply traction loss
            Dim tf_tl As New RotateTransform3D(New AxisAngleRotation3D(New Vector3D(0, 0, 1), Me.tl), New Point3D(Me.tl_dx, 0, 0))

            ' apply surge, sway, heave

            Dim sway_surge_vector_len As Single = Math.Sqrt(Me.sway ^ 2 + Me.surge ^ 2)
            Dim sway_surge_init_angle As Single = Math.Atan2(Me.surge, Me.sway)
            Dim sway_surge_resulting_angle As Single = sway_surge_init_angle - ToRadians(Me.tl)

            Dim sway_surge_dx As Single = sway_surge_vector_len * Math.Sin(sway_surge_resulting_angle)
            Dim sway_surge_dy As Single = sway_surge_vector_len * Math.Cos(sway_surge_resulting_angle)


            Dim tf_surge_sway_heave As New TranslateTransform3D(sway_surge_dx, sway_surge_dy, Me.heave)



            Dim ap As Point3D()
            ReDim ap(5)
            Dim ap_i As Integer = 0
            For Each a As Actuator In Me.actuators.Values
                ap(ap_i) = a.attachment_point
                ap_i += 1
            Next a

            tf_roll.Transform(ap)
            tf_pitch.Transform(ap)
            tf_tl.Transform(ap)
            tf_surge_sway_heave.Transform(ap)

            ap_i = 0
            For Each a As Actuator In Me.actuators.Values
                a.attachment_point = ap(ap_i)
                ap_i += 1
            Next a


            ' all coordinates are updated for the global frame

            ' need to convert to motor frames



            For Each act As Actuator In Me.actuators.Values

                Dim pt_in_m As Point3D = act.AttachmentPointInMotorFrame



                Dim d As Single = Math.Sqrt(pt_in_m.Y ^ 2 + pt_in_m.Z ^ 2)


                Dim r1 As Single = Math.Sqrt(act.rod_length ^ 2 - pt_in_m.X ^ 2)
                Dim r2 As Single = act.lever_length

                If d > r1 + r2 OrElse d < Math.Abs(r1 - r2) Then
                    ' no solutions
                    Debug.Print("no solutions for " & act.name)


                Else

                    Dim int1 As PointF
                    Dim int2 As PointF

                    Call FindCircleCircleIntersections(0, 0, r2, pt_in_m.Y, pt_in_m.Z, r1, int1, int2)


                    Dim shaft_angle1 As Single = Math.Atan(int1.Y / int1.X) + Math.PI
                    If int1.Y > 0 And shaft_angle1 < Math.PI Then
                        shaft_angle1 += Math.PI
                    End If



                    Dim shaft_angle2 As Single = Math.Atan(int2.Y / int2.X)
                    If int2.Y < 0 And shaft_angle2 < 0 Then
                        shaft_angle2 += Math.PI
                    End If


                    act.SetLeverAngle(ToDegrees(shaft_angle1), ToDegrees(shaft_angle2))

                End If

            Next act


        End Sub


        Private Function FindCircleCircleIntersections(ByVal cx0 As Single, ByVal cy0 As Single, ByVal radius0 As Single, ByVal cx1 As Single, ByVal cy1 As Single, ByVal radius1 As Single, ByRef intersection1 As PointF, ByRef intersection2 As PointF) As Integer
            ' Find the distance between the centers.
            Dim dx As Single = cx0 - cx1
            Dim dy As Single = cy0 - cy1
            Dim dist As Double = Math.Sqrt(dx * dx + dy * dy)

            ' See how many solutions there are.
            If (dist > radius0 + radius1) Then
                ' No solutions, the circles are too far apart.
                intersection1 = New PointF(Single.NaN, Single.NaN)
                intersection2 = New PointF(Single.NaN, Single.NaN)
                Return 0
            ElseIf (dist < Math.Abs(radius0 - radius1)) Then
                ' No solutions, one circle contains the other.
                intersection1 = New PointF(Single.NaN, Single.NaN)
                intersection2 = New PointF(Single.NaN, Single.NaN)
                Return 0
            ElseIf ((dist = 0) AndAlso (radius0 = radius1)) Then
                ' No solutions, the circles coincide.
                intersection1 = New PointF(Single.NaN, Single.NaN)
                intersection2 = New PointF(Single.NaN, Single.NaN)
                Return 0
            Else
                ' Find a and h.
                Dim a As Double = (radius0 * radius0 -
                        radius1 * radius1 + dist * dist) / (2 *
                            dist)
                Dim h As Double = Math.Sqrt(radius0 * radius0 - a *
                    a)

                ' Find P2.
                Dim cx2 As Double = cx0 + a * (cx1 - cx0) / dist
                Dim cy2 As Double = cy0 + a * (cy1 - cy0) / dist

                ' Get the points P3.
                intersection1 = New PointF(
                    CSng(cx2 + h * (cy1 - cy0) / dist),
                    CSng(cy2 - h * (cx1 - cx0) / dist))
                intersection2 = New PointF(
                    CSng(cx2 - h * (cy1 - cy0) / dist),
                    CSng(cy2 + h * (cx1 - cx0) / dist))

                ' See if we have 1 or 2 solutions.
                If (dist = radius0 + radius1) Then Return 1
                Return 2
            End If
        End Function




    End Class

    Public Class Actuator

        Public name As String

        ' point where rod is attached to the cockpit, relative to cockpit origin
        Public attachment_x As Integer ' coordinate on front-back axis  positive = forward
        Public attachment_y As Integer ' cooridnate on left-right axis  positive = left
        Public attachment_z As Integer ' coordinate on up-down axis     negative = up


        ' lever origin, relative to base origin
        Public motor_x As Integer ' coordinate on front-back axis  positive = forward
        Public motor_y As Integer ' cooridnate on left-right axis  positive = left
        Public motor_z As Integer ' coordinate on up-down axis     negative = up

        Public motor_angle As Integer

        Public lever_length As Integer

        Public rod_length As Integer


        Public attachment_point As Point3D
        Public motor_axis_x As Point3D
        Public motor_axis_y As Point3D
        Public motor_origin As Point3D

        Public lever_angle As Double

        Public min_angle As Integer
        Public max_angle As Integer







        Public Sub New(ByVal name As String,
                       ByVal attachment_x As Integer,
                       ByVal attachment_y As Integer,
                       ByVal attachment_z As Integer,
                       ByVal motor_x As Integer,
                       ByVal motor_y As Integer,
                       ByVal motor_z As Integer,
                       ByVal motor_angle As Integer,
                       ByVal lever_length As Integer,
                       ByVal rod_length As Integer,
                       ByVal min_angle As Integer,
                       ByVal max_angle As Integer)


            Me.attachment_x = attachment_x
            Me.attachment_y = attachment_y
            Me.attachment_z = attachment_z




            Call AttachmentPointReset()

            Me.motor_angle = motor_angle
            Me.motor_x = motor_x
            Me.motor_y = motor_y
            Me.motor_z = motor_z

            Me.lever_length = lever_length
            Me.rod_length = rod_length

            Me.name = name


            Dim tf_m_rot As New RotateTransform3D(New AxisAngleRotation3D(New Vector3D(0, 0, 1), motor_angle))
            Dim tf_m_trans As New TranslateTransform3D(motor_x, motor_y, 0)

            motor_origin = tf_m_trans.Transform(tf_m_rot.Transform(New Point3D(0, 0, 0)))
            motor_axis_x = tf_m_trans.Transform(tf_m_rot.Transform(New Point3D(100, 0, 0)))
            motor_axis_y = tf_m_trans.Transform(tf_m_rot.Transform(New Point3D(0, 100, 0)))

            Me.min_angle = min_angle
            Me.max_angle = max_angle


        End Sub


        Public Sub AttachmentPointReset()
            attachment_point = New Point3D(Me.attachment_x, Me.attachment_y, Me.attachment_z)
        End Sub

        Public ReadOnly Property AttachmentPointInMotorFrame As Point3D
            Get


                Dim tf_m_rot As New RotateTransform3D(New AxisAngleRotation3D(New Vector3D(0, 0, 1), -motor_angle))
                Dim tf_m_trans As New TranslateTransform3D(-motor_x, -motor_y, 0)

                Dim pt_in_m As Point3D = tf_m_trans.Transform(attachment_point)
                pt_in_m = tf_m_rot.Transform(pt_in_m)

                Return pt_in_m


            End Get
        End Property


        Public Sub SetLeverAngle(ByVal angle1 As Single, ByVal angle2 As Single)

            Dim angle As Single = Single.NaN

            If angle1 <= max_angle AndAlso angle1 >= min_angle Then
                angle = angle1
            ElseIf angle2 <= max_angle AndAlso angle2 >= min_angle Then
                angle = angle2
            End If

            If angle <> Single.NaN Then
                lever_angle = angle
            End If



        End Sub


        Public ReadOnly Property Target As Integer
            Get
                ' assume that target for neutral angle is 5000

                Dim neutral_angle As Single
                Dim offset As Single

                If min_angle >= 90 Then
                    neutral_angle = 180
                    offset = (Me.lever_angle - neutral_angle) / 180
                Else
                    neutral_angle = 0
                    offset = (neutral_angle - Me.lever_angle) / 180
                End If



                'Return CInt(5000 + offset * 5000)

                ' Debug.Print(CInt(5000 + offset * 5000))



                Return CInt(5000 + offset * 5000)



            End Get
        End Property











    End Class















End Module
